#!/usr/bin/python

# Copyright 2016 Bela Genge
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# OptimFlow version number
__version__ = "0.1"

from mod_node import ModNode
from mod_link import ModLink
from mod_flow import ModFlow
import mod_network
from mod_models import ModIFIPNetworkingModel
from mod_setup import ModSetup
from mod_controller import ModController
from mod_solver import ModSolver

import mod_emswitch
from mod_emswitch import ModEMSwitch

from pox.boot import boot_emulated_switch
from time import sleep

from mod_network import TransientNetwork

# Initialize the emulated controller's data as a global object
mod_emswitch.of_emsw = mod_emswitch.initialize()

def buildTopology():

    # Read setup file
    setup = ModSetup()
    setup.readSetup()

    # Get SCIP solver's path
    scippath = setup.getSCIPPath()

    # Create a new model
    mod_network.of_network = ModIFIPNetworkingModel('Network')
    mod_network.of_network.setSecurProp(setup.getSecurProp())
    mod_network.of_network.setBackupLinks(setup.getBackupLinks())

    # Add nodes and their properties
    for n in setup.getNodes():
        nwn = ModNode(n)
        nwn.setProp(setup.getNodeProp(n))
        mod_network.of_network.addNode(nwn)

    # Add links and their properties
    for ll in setup.getLinks():
        ln = ModLink(ll[0])
        ln.setProp(setup.getLinkProp(ll[0]))
        ln.setSecurProp(setup.getLinkSecurProp(ll[0]))
        mod_network.of_network.addLink(ln)

    # Add flows and their properties
    for ff in setup.getFlows():
        fl = ModFlow(ff[0])
        fl.setProp(setup.getFlowProp(ff[0]))
        fl.setSecurProp(setup.getFlowSecurProp(ff[0]))
        mod_network.of_network.addFlow(fl)

    # Configure the properties of the emulated switch
    mod_emswitch.of_emsw.addProp(ModEMSwitch.SWID, setup.getEmSwitchID())
    mod_emswitch.of_emsw.addProp(ModEMSwitch.CTHOST, setup.getEmSwitchCtHost())
    mod_emswitch.of_emsw.addProp(ModEMSwitch.CTPORT, setup.getEmSwitchCtPort())
    mod_emswitch.of_emsw.addProp(ModEMSwitch.INUSE, setup.getEmSwitchUsed())
    swports = setup.getEmSwitchPorts()
    portidx = 1 # This will hold the port index
    for sp_id in swports:
        prop = setup.getEmSwitchProp(sp_id)
        mod_emswitch.of_emsw.ports[portidx] = prop
        portidx = portidx + 1

    return scippath


if __name__ == '__main__':

    print "OptimFlow version: ", __version__

    # Build network topology from setup file
    scippath = buildTopology()
    mod_network.of_network.printNetwork()

    # Launch emulated POX switch
    if mod_emswitch.of_emsw.getProp(ModEMSwitch.INUSE) == 'Yes':
        boot_emulated_switch(controlPort=mod_emswitch.of_emsw.getProp(ModEMSwitch.CTPORT),
            swDpid=mod_emswitch.of_emsw.getProp(ModEMSwitch.SWID),
            swPorts=len(ModEMSwitch.ports))

    # Generate LP file and solve solution - this will create our first working topology
    mod_network.of_network.generateLPFile()
    solver = ModSolver(scippath)
    sol = solver.solve()
    print "*** Solution: ", sol
    print "*** Solution mapping: "
    map = mod_network.of_network.getSolutionMapping(sol)
    print map

    # Build in-memory map of network and push static flows
    mc = ModController()

    # We push the first set of flows irrespectivelly of the network status
    mc.reconfigNetwork(map)

    # Run the main loop that sequentially check the status of the network.
    # In the case of a change in the topology run the optimization algorithm once again and push the static flows.
    while True:

        # Read the state of the network from controller
        (activePorts, inactivePorts) = mc.getNetworkPortStatus()        

        # Update the status of the in-memory network map with the state received from switches
        statusChange = mod_network.of_network.updateNetworkStatusEx(activePorts, inactivePorts)

        # We also need to check if there has been a change in the configuration of static flows pushed in the hierarchical model
        statusChangeEx = mod_network.of_network.performHierarchicalJobs()

        # In the case there is a change of status solve the optimization problem and push static routes
        if statusChange == True or statusChangeEx:

            # Solve new problem
            print "*** Status changed (local={0},hierarchical={1}), running reconfiguration procedure ***".format(statusChange, statusChangeEx)
            print "Active ports: ", activePorts
            print "Inactive ports: ", inactivePorts
            mod_network.of_network.generateLPFile()
            sol = solver.solve()
            print "*** Solution: ", sol
            print "*** Solution mapping: "
            map = mod_network.of_network.getSolutionMapping(sol)
            print map

            # Push routes
            mc.reconfigNetwork(map)
        else:
            print "\t... state not changed."

        # Delay to avoid fast reconfigurations
        sleep(0.1)
