# Copyright 2016 Bela Genge
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
OptimalFlow/mod_setup/ModController
Provides communication functionality with a controller - Floodlight
@author: Bela Genge
'''

import mod_network
from mod_network import ModNetwork
from mod_node import ModNode
from mod_link import ModLink
from mod_flow import ModFlow
import pycurl, json
import StringIO

class ModController():
    '''
    '''

    __endurl = "/flow/json" # We ask per-flow stat, we can also ask per-port stat

    def __init__(self, host = 'localhost', port = 8080):
        '''
        '''
        self.__starturl = 'http://' + host + ':' + str(port) + '/wm/core/switch/'
        self.__statflowurl = 'http://' + host + ':' + str(port) + '/wm/staticflowpusher/json'
        self.__linkstaturl = 'http://' + host + ':' + str(port) + '/wm/topology/links/json'
        self.__statportsurl = 'http://' + host + ':' + str(port) + '/wm/core/switch/all/features/json'

        # Initialize configuration
        fnames = mod_network.of_network.getFlowNames()
        self.__config = {}
        for f in fnames:
            self.__config[f] = []

        # Initialize connection object with Floodlight
        self.__curl = pycurl.Curl()

        # Order the flows in descending order according to their ALPHA property
        self.__orderedFlows = []
        print "FLOWS: "
        print mod_network.of_network._flows
        for (f,fobj) in mod_network.of_network._flows.iteritems():
            pr = int(mod_network.of_network._flows[f].getProp(ModFlow.ALPHA))

            idx = -1
            for idx1 in range(0, len(self.__orderedFlows)):
                pr1 = int(mod_network.of_network._flows[self.__orderedFlows[idx1]].getProp(ModFlow.ALPHA))

                if pr1 < pr:
                    idx = idx1
                    break

            if idx < 0:
                self.__orderedFlows.append(f)
            else:
                self.__orderedFlows.insert(idx, f)

        print "==== ORDERED SET OF FLOWS ====="
        print self.__orderedFlows


    def readNodeStatus(self, nodeID):
        '''
        Read the port/flow status from controller
        '''
        stat = self._readDataFromController(nodeID)
        print stat


    def _readDataFromController(self, nodeID):
        inbuffer = StringIO.StringIO()

        # Setup curl
        #curl = pycurl.Curl()
        self.__curl.reset()
        self.__curl.setopt(pycurl.WRITEFUNCTION, inbuffer.write)
        self.__curl.setopt(pycurl.URL, self.__starturl + nodeID + ModController.__endurl)
        self.__curl.setopt(pycurl.FOLLOWLOCATION, 1)

        # Run curl
        self.__curl.perform()

        # Get result
        return json.loads(inbuffer.getvalue())


    def _pushStaticFlow(self, nodeID, flowUniqueID, src_ip, dst_ip, input, output, protocol):

        # Build static flow
        flow = {
            "switch": nodeID,
            "name": flowUniqueID, # Must be a unique system-level ID (across all switches)
            "cookie": "0",
            "priority": "32765",
            "active": "true",
            "in_port": input,
            "actions": "output="+str(output)
        }

        if protocol == ModFlow.P_ARP:
            flow['eth_type'] = '0x806'
            flow['arp_tpa'] = dst_ip
        else:# At this point we distinguish only between ARP and IP protocols
            flow['eth_type'] = '0x800'
            flow['ipv4_src'] = src_ip
            flow['ipv4_dst'] = dst_ip
            
        print "------- Pushing static flow ---------"
        print flow

        # Setup curl
        #curl = pycurl.Curl()
        self.__curl.reset()
        self.__curl.setopt(pycurl.POST, 1)
        self.__curl.setopt(pycurl.POSTFIELDS, json.dumps(flow))
        self.__curl.setopt(pycurl.URL, self.__statflowurl)
        self.__curl.setopt(pycurl.FOLLOWLOCATION, 1)

        # Run curl
        self.__curl.perform()
        print "\n-------------------------------------"


    def _deleteStaticFlow(self, nodeID, flowUniqueID):

        # Build static flow
        flow = {
#            "switch": nodeID,
            "name": flowUniqueID, # Must be a unique system-level ID (across all switches)
        }

        # Setup curl
        #curl = pycurl.Curl()
        self.__curl.reset()
        self.__curl.setopt(pycurl.CUSTOMREQUEST, 'DELETE')
        self.__curl.setopt(pycurl.POSTFIELDS, json.dumps(flow))
        self.__curl.setopt(pycurl.URL, self.__statflowurl)
        self.__curl.setopt(pycurl.FOLLOWLOCATION, 1)

        # Run curl
        self.__curl.perform()
        print "\n-------------------------------------"


    def _isNodeRouteChanged(self, (j,l,b), f, route, direction):

        # If this function is called, then (j,l,b) is in the new and old route as well.
        # We need to check if the other end of j has changed.
        idx = self.__config[f].index((j,l,b))
        idx1 = route.index((j,l,b))

        # When in the new route we do not have any nodes before, while in the old route we still have something, or vice-versa
        if (idx1 == 0 and idx > 0) or (idx1 > 0 and idx == 0):
            return True

        # Otherwise, both are greater than 0, we can safely check the previous node and links
        (jx,lx,bx) = self.__config[f][idx-1]
        (jx1,lx1,bx1) = route[idx1-1]
        if jx != jx1 or bx != bx1:
            return True
        
        return False


    def _pushStaticToNode(self, f, (j,l,b), route, reverse = False):

        flow = mod_network.of_network.getFlow(f)

        # Get switch ID
        swid = mod_network.of_network.getNode(j).getProp(ModNode.MAC)

        # Get flow protocol
        protocol = flow.getProp(ModFlow.PROTO)

        dirstr = 'orig' # Hardcoded string used to build unique flow ID, it can be anything 
        input = -1
        output = -1
        src_ip = ''
        dst_ip = ''
        if reverse == False:
            src_ip = flow.getProp(ModFlow.ACCIP)
            dst_ip = flow.getProp(ModFlow.EGGIP)

            # Calculate output port
            if route.index((j,l,b)) >= len(route)-1:
                output = int(flow.getProp(ModFlow.EGGPORT))
            else:
                link = mod_network.of_network.getLink(j,l,b)

                # Get the output port (next hop)
                output = -1
                if link.getProp(ModLink.N1) == j:
                    output = link.getProp(ModLink.P1)
                else:
                    output = link.getProp(ModLink.P2)

            # Calculate input port
            if route.index((j,l,b)) <= 0:
                input = int(flow.getProp(ModFlow.ACCPORT))
            else:
                # Get the link with the previous node
                (j1,l1,b1) = route[route.index((j,l,b))-1]
                link = mod_network.of_network.getLink(j1,l1,b1)

                # Get the input port
                input = -1
                if link.getProp(ModLink.N1) == j:
                    input = link.getProp(ModLink.P1)
                else:
                    input = link.getProp(ModLink.P2)
        else:
            src_ip = flow.getProp(ModFlow.EGGIP)
            dst_ip = flow.getProp(ModFlow.ACCIP)
            dirstr = 'reverse' # Hardcoded string used to build unique flow ID, can be anything

            # Calculate output port
            if route.index((j,l,b)) == 0:
                output = int(flow.getProp(ModFlow.ACCPORT))
            else:
                # Get the next node's ID so that we can identify the link between the two nodes
                (j1,l1,b1) = route[route.index((j,l,b))-1]
                link = mod_network.of_network.getLink(j1,l1,b1)

                # Get the output port (next hop)
                output = -1
                if link.getProp(ModLink.N1) == j:
                    output = link.getProp(ModLink.P1)
                else:
                    output = link.getProp(ModLink.P2)

            # Calculate input port
            if route.index((j,l,b)) >= len(route)-1:
                input = int(flow.getProp(ModFlow.EGGPORT))
            else:
                # Get the next node's ID
                link = mod_network.of_network.getLink(j,l,b)

                # Get the input port (next hop)
                input = -1
                if link.getProp(ModLink.N1) == j:
                    input = link.getProp(ModLink.P1)
                else:
                    input = link.getProp(ModLink.P2)

        # Finally, push static flow
        self._pushStaticFlow(swid, f+"_"+swid+"_"+dirstr, src_ip, dst_ip, input, output, protocol)


    def _pushRoute(self, f, route, direction):

        for (j,l,b) in route:

            shouldAdd = False
            if (j,l,b) not in self.__config[f]:
                # The simplest case - when there is a new node in the route
                shouldAdd = True
            else:
                shouldAdd = self._isNodeRouteChanged((j,l,b), f, route, direction) # However, most often we have changes in the output ports

            if route.index((j,l,b)) <= 0:
                if mod_network.of_network.getFlow(f).accChanged:
                    shouldAdd = True
            if route.index((j,l,b)) >= (len(route)-1):
                if mod_network.of_network.getFlow(f).eggChanged:
                    shouldAdd = True

            if shouldAdd == True:
                self._pushStaticToNode(f, (j,l,b), route)
                if direction == ModFlow.D_BIDIR:
                    self._pushStaticToNode(f, (j,l,b), route, True)


    def _deleteStaticToNode(self, f, nodeID, reverse = False):

        flow = mod_network.of_network.getFlow(f)

        # Get switch ID
        swid = mod_network.of_network.getNode(nodeID).getProp(ModNode.MAC)

        dirstr = 'orig' # Hardcoded string used to build unique flow ID, can be anything 
        if reverse == True:
            dirstr = 'reverse' # Hardcoded string used to build unique flow ID, can be anything 

        self._deleteStaticFlow(swid, f+"_"+swid+"_"+dirstr)


    def _deleteRoute(self, f, route, direction):

        # It is actually not enough to check for nodes, we also need to check if the output link has changed
        for (j,l,b) in self.__config[f]:

            shouldDelete = False
            # The most obvious - if the link is no longer part of the route, we need to delete it
            if (j,l,b) not in route:
                shouldDelete = True
            else:
                # However, usually, we have a change in the output port
                shouldDelete = self._isNodeRouteChanged((j,l,b), f, route, direction)
                if shouldDelete == True:
                    print "We detected a change in flow: ", f, shouldDelete, (j,l,b), route, self.__config[f]

            if self.__config[f].index((j,l,b)) <= 0:
                if mod_network.of_network.getFlow(f).accChanged:
                    shouldDelete = True
            if self.__config[f].index((j,l,b)) >= (len(self.__config[f])-1):
                if mod_network.of_network.getFlow(f).eggChanged:
                    shouldDelete = True

            if shouldDelete == True:
                # Delete static route for flow f in node n
                self._deleteStaticToNode(f, j)
                if direction == ModFlow.D_BIDIR:
                    self._deleteStaticToNode(f, j, True)


    def reconfigNetwork(self, routes):
        '''
        Probably the most important method: reconfigures the network by deleting and pushing static routes.
        '''

        # Check the differences between the old configuration and the new one
        for f in self.__orderedFlows:
            route = None
            # Search for this flow and its route
            for (f1,route1) in routes.iteritems():
                if f == f1:
                    route = route1
                    break

            if None == route:
                continue

            # Get direction: one direction or bidirectional flow
            direction = mod_network.of_network.getFlow(f).getProp(ModFlow.DIR)

            # First, delete the flow configurations that are not in the new route
            # It is actually not enough to check for nodes, we also need to check if the output link has changed
            self._deleteRoute(f, route, direction)

            # Next, push flows
            self._pushRoute(f, route, direction)

            # Reset change flags
            mod_network.of_network.getFlow(f).accChanged = False
            mod_network.of_network.getFlow(f).eggChanged = False

        self.__config = routes            


    def getNetworkStatus(self):
        '''
        Reads the controller for Returns the list of active and inactive links between CSs (switches).
        !Important! The network topology is built by using the LLDP protocol. This means, however, that
        in the hierarchical case of OptimFlow we cannot really use it...
        '''

        inbuffer = StringIO.StringIO()

        # Setup curl
        #curl = pycurl.Curl()
        self.__curl.reset()
        self.__curl.setopt(pycurl.WRITEFUNCTION, inbuffer.write)
        self.__curl.setopt(pycurl.URL, self.__linkstaturl )
        self.__curl.setopt(pycurl.FOLLOWLOCATION, 1)

        # Run curl
        self.__curl.perform()

        # Get result
        res = json.loads(inbuffer.getvalue())

        # Links that are in the list are active. The rest are inactive
        activeLinks = []
        for link in res:
            sw1 = link['src-switch']
            sw2 = link['dst-switch']

            # Find their IDs
            id1 = mod_network.of_network.getNodeMAC(sw1)
            id2 = mod_network.of_network.getNodeMAC(sw2)

            if id1 == None:
                print "ERROR: unable to find internal ID for switch: ", sw1
                continue
            if id2 == None:
                print "ERROR: unable to find internal ID for switch: ", sw2
                continue

            activeLinks.append((id1, id2))

            # Check if this link is 

        # Build list with inactive links
        inactiveLinks = []
        for ((cs1,cs2),link) in mod_network.of_network._links.iteritems():

            # Check if this link is active or not
            isActive = False
            for (id1,id2) in activeLinks:
                if ((id1 == cs1 and id2 == cs2) or (id1 == cs2 and id2 == cs1)):
                    isActive = True
                    break

            if isActive == False:
                inactiveLinks.append((cs1,cs2))
        
        return (activeLinks, inactiveLinks)


    def getNetworkPortStatus(self):
        '''
        Builds the state of the network from the point of view of port status. It does not require LLDP to be active,
        so it can be used in the case of hierarchical OptimFlow controllers.
        '''
        inbuffer = StringIO.StringIO()

        # Setup curl
        #curl = pycurl.Curl()
        self.__curl.reset()
        self.__curl.setopt(pycurl.WRITEFUNCTION, inbuffer.write)
        self.__curl.setopt(pycurl.URL, self.__statportsurl )
        self.__curl.setopt(pycurl.FOLLOWLOCATION, 1)

        # Run curl
        self.__curl.perform()

        # Get result
        res = json.loads(inbuffer.getvalue())

        # The result is a dictionary of (switchID,features) pairs
        # We need to construct a dictionary of active and inactive ports on switches (switchID,[list of active/inactive switches])
        activePorts = {}
        inactivePorts = {}
        for (sw,feat) in res.iteritems():

            # Sometimes we get some error messages for unknown switches... not sure what t$
            if 'portDesc' not in feat:
                continue

            for ff in feat['portDesc']:
                if ff['state'] == '0':
                    # Active port
                    if sw not in activePorts:
                        activePorts[sw] = [ff['portNumber']]
                    else:
                        activePorts[sw].append(ff['portNumber'])
                else:
                    # Inactive port
                    if sw not in inactivePorts:
                        inactivePorts[sw] = [ff['portNumber']]
                    else:
                        inactivePorts[sw].append(ff['portNumber'])

        return (activePorts, inactivePorts)

