# Copyright 2016 Bela Genge
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
OptimalFlow/mod_models/ModBaseModel

@author: Bela Genge
'''

from mod_network import ModNetwork
from mod_flow import ModFlow
from mod_link import ModLink

class ModBaseModel(ModNetwork):
    '''
    '''

    def __init__(self, name):
        '''
        Constructor
        '''
        super(ModBaseModel, self).__init__(name)

        self._tmap = {}
        self._wAmap = {}
        self._wEmap = {}
        self._omap = {}
        self._cctr = 0 # Constraint counter
        self.__lpfname = './OptimalFlow.lp'


    def generateLPFile(self, fname=None):
        '''
        Function used to generate the LP model file from the network model
        '''

        # Create/open the file for writing
        if fname is None:
            lpf = open(self.__lpfname, "w")
        else:
            lpf = open(fname, "w")

        # Write header
        lpf.write("\\ENCODING=ISO-8859-1\n\\Problem name: CPLEX-API\n\n")

        # Generate t_jlbi variable indexes
        if len(self._tmap) <= 0:
            self._gen_varMapping()

        # Write objective
        self._writeObjective(lpf)

        # Write constraints
        self._writeConstraints(lpf)

        # Write bounds and binary variables
        self._writeCBoundsBinaries(lpf)

        # End LP file
        lpf.write("End\n")

        self._printTmap()

        # Close file
        lpf.close()


    def getSolutionMapping(self, vars):

        # Get first mapping for the solution variables
        fmap_t = {}
        fmap_wA = {}
        fmap_wE = {}
        fmap_o = []
        for v in vars:

            # t-variables
            if v <= len(self._tmap):
                (j,l,b,i) = self._get_t_varMapping(v)
                if i not in fmap_t:
                    fmap_t[i] = [(j,l,b)]
                else:
                    fmap_t[i].append((j,l,b))
            elif v <= len(self._tmap) + len(self._wAmap):
                (i,j) = self._get_wA_varMapping(v)
                if i not in fmap_wA:
                    fmap_wA[i] = j
            elif v <= len(self._tmap) + len(self._wAmap) + len(self._wEmap):
                (i,j) = self._get_wE_varMapping(v)
                if i not in fmap_wE:
                    fmap_wE[i] = j
            else:
                i = self._get_o_varMapping(v)
                if i is not None:
                    fmap_o.append(i)

        # Next, order the nodes in the t-mapping so that we get the path
        fmap1 = {}
        for (i,map) in fmap_t.iteritems():
            fmap1[i] = []

            # This flow should be completely removed
            if i in fmap_o:
                # Actually, this test is useless, but we keep it (paranoid reasons), since fmap_t will not include the flows to be removed!
                continue

            # Start from Access node and step through the list until we reach Egress node
            n = fmap_wA[i]
            ne = fmap_wE[i]
            while n != ne:
                for (j,l,b) in map:
                    if n == j:
                        fmap1[i].append((j,l,b))
                        n = l
            #if ne not in fmap1[i]:
            fmap1[i].append((ne,ne,'NA')) # For the last node we do not have a specific link

        # Now add the flows that have been removed (with an empty route)
        for i in fmap_o:
            if i not in fmap1:
                fmap1[i] = []

        return fmap1


    def _printTmap(self):
        print "*** t-Variable mapping: "
        for ctr in range(1,len(self._nodes)*len(self._nodes)*len(self._backup)*len(self._flows)+1):
            for (j,l,b,i),idx in self._tmap.iteritems():
                if idx == ctr:
                    print "x", ctr, ": ", j, l, b, i
                    break

        st = len(self._nodes)*len(self._nodes)*len(self._backup)*len(self._flows)
        print "*** wA-Variable mapping: "
        for ctr in range(1,len(self._flows)*len(self._nodes)+1):
            for (i,j),idx in self._wAmap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", i, j
                    break

        st += len(self._nodes)*len(self._flows)
        print "*** wE-Variable mapping: "
        for ctr in range(1,len(self._flows)*len(self._nodes)+1):
            for (j,i),idx in self._wEmap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", j, i
                    break

        st += len(self._nodes)*len(self._flows)
        print "*** o-Variable mapping: "
        for ctr in range(1,len(self._flows)+1):
            for i,idx in self._omap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", i
                    break


    def _get_t_varMapping(self, var):

        for j,jobj in self._nodes.iteritems():
            for l,lobj in self._nodes.iteritems():
                for b in self._backup:
                    for i,iobj in self._flows.iteritems():
                        if self._tmap[(j,l,b,i)] == var:
                            return (j,l,b,i)
        return None


    def _get_wA_varMapping(self, var):

        for i,iobj in self._flows.iteritems():
            for j,jobj in self._nodes.iteritems():
                if self._wAmap[(i,j)] == var:
                    return (i,j)

        return None


    def _get_wE_varMapping(self, var):
        
        for j,jobj in self._nodes.iteritems():
            for i,iobj in self._flows.iteritems():
                if self._wEmap[(j,i)] == var:
                    return (i,j)

        return None


    def _get_o_varMapping(self, var):

        for i,iobj in self._flows.iteritems():
            if self._omap[i] == var:
                return i

        return None


    def _gen_varMapping(self):
        # Generate the mapping between network names and variable names
        self._tmap = {}
        self._wAmap = {}
        self._wEmap = {}
        self._omap = {}
        vidx = 1

        # Generate mapping for t
        for j,jobj in self._nodes.iteritems():
            for l,lobj in self._nodes.iteritems():
                for b in self._backup:
                    for i,iobj in self._flows.iteritems():
                        self._tmap[(j,l,b,i)] = vidx
                        vidx += 1

        # Generate mapping for w_A
        for i,iobj in self._flows.iteritems():
            for j,jobj in self._nodes.iteritems():
                self._wAmap[(i,j)] = vidx
                vidx += 1

        # Generate mapping for w_E
        for j,jobj in self._nodes.iteritems():
            for i,iobj in self._flows.iteritems():
                self._wEmap[(j,i)] = vidx
                vidx += 1

        # Generate mapping for o
        for i,iobj in self._flows.iteritems():
            self._omap[i] = vidx
            vidx += 1


    def _writeObjective(self, lpf):
        '''
        Write the objective function.
        '''
        pass


    def _writeConstraints(self, lpf):
        '''
        Write constraints.
        '''

        self._cctr = 0

        # Write flow conservation constraints
        self._writeCFlowConservation(lpf)

        # Write link capacity constraints
        self._writeCLinkCapacity(lpf)


    def _writeCFlowConservation(self, lpf):
        '''
        Write flow conservation constraints: w_ij^A - w_ji^E - sum_lb(t_jlbi - t_ljbi) = 0
        that is: - sum_lb(t_jlbi - t_ljbi) = w_ji^E - w_ij^A 
        '''

        lpf.write("Subject To\n")

        for j,jobj in self._nodes.iteritems():
            for i,iobj in self._flows.iteritems():

                # Prepare list of variables for ordering - so it is easier to read
                # Make sure we do not lose the sign
                tvars = []
                tvarsign = {}
                for l,lobj in self._nodes.iteritems():

                    # In the case of j = l we skip and go to the next variable
                    if j == l:
                        continue

                    for b in self._backup:

                        # -(t_jlbi - tljbi)
                        tvars.append(self._tmap[(j,l,b,i)])
                        tvarsign[self._tmap[(j,l,b,i)]] = -1

                        tvars.append(self._tmap[(l,j,b,i)])
                        tvarsign[self._tmap[(l,j,b,i)]] = 1

                # Perform the ordering
                tvars.sort()

                # Write the constraint
                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ":  ")

                first = True
                for x in tvars:
                    if tvarsign[x] == -1:
                        lpf.write("- ")
                    elif first == False:
                        lpf.write("+ ")
                    lpf.write("x" + str(x) + " ")
                    first = False

                lpf.write("+ x")
                lpf.write(str(self._wAmap[(i,j)]))
                lpf.write(" - x")
                lpf.write(str(self._wEmap[(j,i)]))
                lpf.write(" = 0\n")


    def _writeCLinkCapacity(self, lpf):
        '''
        Write link capacity constraints: sum_i(t_jlbi*d_i) <= u_jlb
        This equation needs to account for bi-directional flows on the same link.
        If we assume that u_jlb = u_ljb, then we can rewrite the above as: sum_i((t_jlbi+t_ljbi)*d_i) <= u_jlb
        '''

        for j,jobj in self._nodes.iteritems():
            for l,lobj in self._nodes.iteritems():
                for b in self._backup:

                    # Write the constraint number
                    self._cctr += 1
                    lpf.write(" c" + str(self._cctr) + ":  ")

                    # Write the sum of variables for t_jlbi*d_i
                    first = True
                    for i,iobj in self._flows.iteritems():
                        if first == True:
                            lpf.write(str(iobj.getProp(ModFlow.DEM)) + " ")
                        else:
                            lpf.write("+ ")
                            lpf.write(str(iobj.getProp(ModFlow.DEM)) + " ")
                        lpf.write("x" + str(self._tmap[(j,l,b,i)]) + " ")
                        first = False

                    # Write the sum of variables for t_ljbi*d_i
                    for i,iobj in self._flows.iteritems():
                        lpf.write("+ ")
                        lpf.write(str(iobj.getProp(ModFlow.DEM)) + " ")
                        lpf.write("x" + str(self._tmap[(l,j,b,i)]) + " ")

                    # Write the inequality
                    lpf.write("<= ")
                    if (j,l,b) in self._links:
                        lpf.write(str(self._links[(j,l,b)].getProp(ModLink.RCAP)))
                    elif (l,j,b) in self._links:
                        lpf.write(str(self._links[(l,j,b)].getProp(ModLink.RCAP)))
                    else:
                        lpf.write("0")
                    lpf.write("\n")


    def _writeCBoundsBinaries(self, lpf):
        '''
        Write bounds and binary variables - all variables actually.
        '''

        varcount = len(self._nodes)*len(self._nodes)*len(self._flows)+ 2*len(self._flows)*len(self._nodes) + len(self._flows)

        lpf.write("Bounds\n")
        # Write bounds for t,w_A,w_E,o
        for ctr in range(1, varcount + 1):
            lpf.write(" 0 <= x" + str(ctr) + " <= 1\n")

        lpf.write("Binaries\n")
        idx = 0
        for ctr in range(1, varcount + 1):
            lpf.write(" x" + str(ctr) + " ")
            # Write max. 17 variables per line - AIMMS style
            idx = idx + 1
            if idx >= 9:
                lpf.write("\n")
                idx = 0
        lpf.write("\n")
