# Copyright 2016 Bela Genge
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
OptimalFlow/mod_models/ModIFIPNetworkingModel
A model used to reconfigurate the network - it is the one included in the IFIP Networking paper.

@author: Bela Genge
'''

from mod_models import ModBaseModel
from mod_flow import ModFlow
from mod_link import ModLink

class ModIFIPNetworkingModel(ModBaseModel):
    '''
    '''

    def __init__(self, name):
        '''
        Constructor
        '''
        super(ModIFIPNetworkingModel, self).__init__(name)


    def _writeObjective(self, lpf):
        '''
        Write the objective function: sum_jl [(1/u_jlb)*sum_i ((t_jlbi + t_ljbi) * d_i)] + sum_i(o_i*_alpha_i)]
        below ocap_jl = 1/u_jlb if u_jlb <> 0 and 0 otherwise
        '''
        super(ModIFIPNetworkingModel, self)._writeObjective(lpf)

        lpf.write("Minimize\n obj:")

        # Initialize ocap_jl
        ocap = {}
        for j,jobj in self._nodes.iteritems():
            for l,lobj in self._nodes.iteritems():
                for b in self._backup:
                    link = self.getLink(j,l,b)
                    if link == None:                
                        ocap[(j,l,b)] = 0
                        continue

                    ocap[(j,l,b)] = 0 if link.getProp(ModLink.RCAP) == 0 else 1.0/link.getProp(ModLink.RCAP)

        # Write the first part: sum_jl [(1/u_jlb)*sum_i (t_jlbi * d_i)]
        ctr = 0
        first = True
        for j,jobj in self._nodes.iteritems():
            for l,lobj in self._nodes.iteritems():
                for b in self._backup:

                    # In case ocap is 0, we do not add the associated t_jlb variable - this will be removed from other constraints as well
                    if ocap[(j,l,b)] == 0:
                        continue

                    for i,iobj in self._flows.iteritems():

                        # Write max 9 variables per line - AIMMS style
                        ctr = ctr + 1
                        if ctr > 9:
                            lpf.write("\n     ")
                            ctr = 0;

                        # Write '-' (negative) summation
                        #lpf.write(" - ")

                        # Write cap*demand value
                        prod = ocap[(j,l,b)]*iobj.getProp(ModFlow.DEM)
                        if first == True:
                            lpf.write(str(prod) + " ")
                            first = False
                        else:
                            lpf.write("+ " + str(prod) + " ")

                        # Write t_jlbi variable name
                        lpf.write("x" + str(self._tmap[(j,l,b,i)]) + " ")

                        # Write again cap*demand value for the second multiplication with t_ljbi
                        lpf.write("+ " + str(prod) + " ")
                        lpf.write("x" + str(self._tmap[(l,j,b,i)]) + " ")
                    
        # Write the second part: sum_i(o_i*_alpha_i)
        for i,iobj in self._flows.iteritems():

            lpf.write(" + " + str(iobj.getProp(ModFlow.ALPHA)) + " x" + str(self._omap[i]))            

        lpf.write("\n")


    def _writeConstraints(self, lpf):
        '''
        Write additional constraints. Make sure we call superclass's version.
        '''

        super(ModIFIPNetworkingModel, self)._writeConstraints(lpf)

        # Write additional constraints
        self._writeAccessConstraints(lpf)
        self._writeEgressConstraints(lpf)
        self._writeSecurityConstraints(lpf)
        self._writePriorityDisconnPenaltyConstraints(lpf)
        self._writeReliabilityConstraints(lpf)


    def _writeAccessConstraints(self, lpf):
        '''
        Write access constraints: w_A(i,j) <= x_A(i,j)
        '''
        
        for i,iobj in self._flows.iteritems():
            for j,jobj in self._nodes.iteritems():
            
                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ": ")
                
                xA_ij = 0
                if iobj.getProp(ModFlow.ACC) == j:
                    xA_ij = 1
                
                lpf.write("x" + str(self._wAmap[(i,j)]) + " <= " + str(xA_ij) + "\n")


    def _writeEgressConstraints(self, lpf):
        '''
        Write egress constraints: w_E(j,i) <= x_E(j,i)
        '''
        
        for j,jobj in self._nodes.iteritems():
            for i,iobj in self._flows.iteritems():
            
                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ": ")
                
                xE_ji = 0
                if iobj.getProp(ModFlow.EGG) == j:
                    xE_ji = 1
                
                lpf.write("x" + str(self._wEmap[(j,i)]) + " <= " + str(xE_ji) + "\n")

    
    def _writeSecurityConstraints(self, lpf):
        '''
        Write security constraints: sum[s, p_is*y_jlbs] >= t_jlbi, that is: - t_jlbi >= - sum[s, p_is*y_jlbs] (in LP file style)
        '''
        for i,iobj in self._flows.iteritems():
            for j,jobj in self._nodes.iteritems():
                for l,lobj in self._nodes.iteritems():

                    if j == l:
                        continue

                    for b in self._backup:

                        self._cctr += 1
                        lpf.write(" c" + str(self._cctr) + ": ")

                        # Write the variable from left hand side
                        lpf.write("- ")
                        lpf.write("x" + str(self._tmap[(j,l,b,i)]))
                        lpf.write(" >= ")

                        # Calculate sum[s, p_is*y_jls]
                        lsec = None
                        if (j,l,b) in self._links:
                            lsec = self._links[(j,l,b)].getSecurProp()
                        elif (l,j,b) in self._links:
                            lsec = self._links[(l,j,b)].getSecurProp()

                        # If we do not have a link between the two nodes, just write 0
                        if lsec == None:
                            lpf.write("0\n")
                            continue

                        ctr = 0
                        for s in self._secur:
                            if s in iobj.getSecurProp() and s in lsec:
                                ctr += 1

                        if ctr == 0:
                            lpf.write("0\n")
                        else:
                            lpf.write("-" + str(ctr) + "\n")


    def _writePriorityDisconnPenaltyConstraints(self, lpf):
        '''
        Write disconnect penalty constraints: sum[j, w_A(i,j)] = 1 - o(i), that is: sum[j, w_A(i,j)] + o(i) = 1 (in LP file style)
        '''

        for i,iobj in self._flows.iteritems():

            self._cctr += 1
            lpf.write(" c" + str(self._cctr) + ": ")

            first = True
            for j,jobj in self._nodes.iteritems():

                if first == True:
                    lpf.write("x" + str(self._wAmap[(i,j)]) + " ")
                    first = False
                else:
                    lpf.write("+ x" + str(self._wAmap[(i,j)]) + " ")

            lpf.write("+ x" + str(self._omap[i]) + " = 1\n")


    def _writeReliabilityConstraints(self, lpf):
        '''
        Write reliability constraints: sum[(j,l), r(j,l) * t(j,l,b,i)] >= q(i)
        '''

        for i,iobj in self._flows.iteritems():

            self._cctr += 1
            lpf.write(" c" + str(self._cctr) + ": ")

            for j,jobs in self._nodes.iteritems():
                for l,lobj in self._nodes.iteritems():

                    if j == l:
                        continue

                    for b in self._backup:
                        if (j,l,b) not in self._tmap:
                            continue

                        # At the moment assume hard-coded reliability values
                        lpf.write(" - 0.00100050033 x"  + str(self._tmap[(j,l,b)]))

            lpf.write(" >= -4.60517018599\n")
