# Copyright 2016 Bela Genge
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
OptimalFlow/mod_network/ModNetwork

@author: Bela Genge
'''

from mod_node import ModNode
from mod_link import ModLink
from mod_flow import ModFlow
import mod_emswitch
from mod_emswitch import ModEMSwitch
import subprocess
import re
import threading
import time
import calendar

of_network = None

class ModNetwork(object):
    '''
    '''

    def __init__(self, name):
        '''
        Constructor
        '''
        super(ModNetwork, self).__init__()

        self.__name = name
        self._nodes = {}
        self._links = {}
        self._flows = {}
        self._secur = [] # The list of security levels
        self._backup = [] # The list of possible (backup) line types

        self.__activePorts = {}
        self.__inactivePorts = {}
        self.__statflowjobs = [] # List of jobs (static flows) that need to be modified in the hierarchical model.
        self.__timelastjob = calendar.timegm(time.gmtime()) # The UTC time in seconds of the last job received from hierarchical controller
        self.__lock = threading.Lock() # Lock used to sync all operations concerning the emulated switch

        
    def setInfo(self, name):
        self.__name = name


    def addNode(self, node):

        # Check if a switch with the same name already exists
        if node.getName() in self._nodes:
            print "WARNING: Node with name {0} already exists. Skipping add node".format(node.getName())
            return False

        # Add node
        self._nodes[node.getName()] = node

        return True


    def getNode(self, node):

        return self._nodes[node]


    def getNodeMAC(self, nodeMAC):

        # Get node based on MAC
        for (nid, n) in self._nodes.iteritems():
            if n.getProp(ModNode.MAC) == nodeMAC:
                return nid

        return None


    def addLink(self, link):

        # Check if the two nodes are in the graph
        '''if (link.getProp(ModLink.N1),link.getProp(ModLink.N2)) not in self._links:
            if (link.getProp(ModLink.N2),link.getProp(ModLink.N1)) in self._links:
                print "WARNING: There is already a link between {0} and {1}. Skipping add link".format(link.getProp(ModLink.N1), link.getProp(ModLink.N2))
                return False
            else:
                pass
        else:
            print "WARNING: There is already a link between {0} and {1}. Skipping add link".format(link.getProp(ModLink.N1), link.getProp(ModLink.N2))
            return False
        '''
            

        # Add link
        self._links[(link.getProp(ModLink.N1),link.getProp(ModLink.N2),link.getProp(ModLink.BKP))] = link

        return True


    def getLink(self, n1, n2, b):

        if (n1,n2,b) in self._links:
            return self._links[(n1,n2,b)]
        elif (n2,n1,b) in self._links:
            return self._links[(n2,n1,b)]

        return None


    def getLinkEx(self, swID, port):
        '''
        Return the link associated to this switch ID and this port number.
        '''

        # Get the node based on MAC
        node = self.getNodeMAC(swID)
        if node == None:
            return None

        # Get the link based on node and port
        for ((n1,n2,b),ll) in self._links.iteritems():
            if (n1 == node) and (ll.getProp(ModLink.P1) == port):
                    return ll
            if (n2 == node) and (ll.getProp(ModLink.P2) == port):
                    return ll

        return None


    def addFlow(self, flow):
        # Check if flow is already present
        if flow.getName() in self._flows:
            print "WARNING: Flow with name {0} already exists. Skipping add flow".format(flow.getName())
            return False

        # Add new flow
        self._flows[flow.getName()] = flow

        return True


    def getFlow(self, fid):
        return self._flows[fid]


    def getFlowEx(self, protocol, src_ip, dst_ip):
        '''
        Return the flow with this protocol and IP address (source or destination).
        '''
        for (fid,f) in self._flows.iteritems():
            if f.getProp(ModFlow.PROTO) == protocol:
                if src_ip == None:  # In the case of the ARP protocol
                    if (f.getProp(ModFlow.ACCIP) == dst_ip) or (f.getProp(ModFlow.EGGIP) == dst_ip):
                        return f
                else:
                    if ((f.getProp(ModFlow.ACCIP) == src_ip and f.getProp(ModFlow.EGGIP) == dst_ip) or
                        (f.getProp(ModFlow.ACCIP) == dst_ip and f.getProp(ModFlow.EGGIP) == src_ip)):
                        return f
        return None


    def getFlowExEx(self, protocol, src_ip, dst_ip):
        '''
        Return the flow with this protocol and IP address (source or destination) - do not check for reverse!!!.
        '''
        for (fid,f) in self._flows.iteritems():
            if f.getProp(ModFlow.PROTO) == protocol:
                if src_ip == None:  # In the case of the ARP protocol
                    if f.getProp(ModFlow.EGGIP) == dst_ip:
                        return f
                else:
                    if (f.getProp(ModFlow.ACCIP) == src_ip and f.getProp(ModFlow.EGGIP) == dst_ip):
                        return f
        return None


    def getFlowNames(self):

        fn = []
        for f in self._flows:
            fn.append(f)

        return fn


    def setSecurProp(self, secur):

        self._secur = secur


    def setBackupLinks(self, backup):

        self._backup = backup


    def printNetwork(self):
        print "*** The list of nodes: "
        for j,jobj in self._nodes.iteritems():
            print j, jobj.getProp(ModNode.MAC)

        print "*** The list of links: "
        for ((n1,n2,b),link) in self._links.iteritems():
            print link.getName(), link.getProp(ModLink.N1), link.getProp(ModLink.N2), link.getProp(ModLink.CAP), link.getProp(ModLink.RCAP), b

        print "*** The list of flows: "
        for i,iobj in self._flows.iteritems():
            print i, iobj.getProp(ModFlow.ACC), iobj.getProp(ModFlow.EGG), iobj.getProp(ModFlow.DEM)

        print "*** The security level: "
        print self._secur

    
    def updateNetworkStatusEx(self, activePorts, inactivePorts):
        '''
        Active links are configured with their predefined capacities, while inactive links are configured with 0 capacity.
        We return True in case the network changed its configuration and False otherwise.
        '''

        # Update active ports
        statusChange = False
        for (swID,ports) in activePorts.iteritems():

            for port in ports:

                # We do not care about 'local' switch ports
                if port == 'local':
                    continue

                link = self.getLinkEx(swID,port)
                if link == None:
                    #print "ERROR: unable to get the link to activate port: ", swID, port
                    # in this case there is not necessarily an error because, for instance there might be links
                    # that are not governed by this controller
                    continue

                # In case there is a change of network topology signal this change.
                # Normally, we should not have a preconfigured capacity of 0 - this would not make sense.
                # However, if for some reason we need a preconfigured capacity of 0, then the algorithm should be changed!
                if link.getProp(ModLink.RCAP) == 0:

                    #print "Statusex changed for: ", id1, id2, link.getProp(ModLink.RCAP)
                    link.addProp(ModLink.RCAP, link.getProp(ModLink.CAP))
                    statusChange = True

        # Update inactive ports
        for (swID,ports) in inactivePorts.iteritems():

            for port in ports:

                # We do not care about 'local' switch ports
                if port == 'local':
                    continue

                link = self.getLinkEx(swID,port)
                if link == None:
                    #print "ERROR: unable to get link to deactivate port: ", swID, port
                    # not necessarily an error - see comment above
                    continue

                # In case there is a change of network topology signal this change.
                # Normally, we should not have a preconfigured capacity of 0 - this would not make sense.
                # However, if for some reason we need a preconfigured capacity of 0, then the algorithm should be changed!
                if link.getProp(ModLink.RCAP) != 0:

                    #print "Status changed for: ", id1, id2
                    link.addProp(ModLink.RCAP, 0)
                    statusChange = True

        # Finally, store the list of active and inactive ports for this emulated switch.
        # This list is used in the hierarchical model to send the status of ports of the emulated controller.
        self.__lock.acquire()

        self.__activePorts = mod_emswitch.of_emsw.getEmPorts(activePorts)
        self.__inactivePorts = mod_emswitch.of_emsw.getEmPorts(inactivePorts)

        self.__lock.release()

        return statusChange


    def getEmSwPortStatus(self):
        active = []
        inactive = []
        self.__lock.acquire()

        active = self.__activePorts
        inactive = self.__inactivePorts

        self.__lock.release()

        return (active, inactive)


    def onAddStaticFlow(self, in_port, out_port, eth_type, src_ip, dst_ip):
        '''
        This method is called when a new static flow is pushed within the hierarchical controller model.
        !Important! The in_port and out_port are those configured in the emulated switch. We therefore
        need to find the associated (real) switch and configure that particular switch as output.
        '''

        # To make sure we a thread-safe, we just push the job in a queue and let the main thread to handle it.
        protocol = ModFlow.P_ARP if str(eth_type)=='0x806' else ModFlow.P_IP
        self.__lock.acquire()

        self.__statflowjobs.append({'TYPE':'ADD', 'INP':in_port, 'OUTP': out_port, 'PROTO': protocol, 'SIP': src_ip, 'DIP': dst_ip})
        self.__timelastjob = calendar.timegm(time.gmtime())
        print "Static flow added to the list of jobs. List content: ", self.__statflowjobs

        self.__lock.release()


    def performHierarchicalJobs(self):
        '''
        This method performs the jobs pushed by the hierarchical controller - it changes configurations in the network structures, but the actual flows are only pushed by the controller!
        '''

        retVal = False
        self.__lock.acquire()

        # First, verify that at least 2 seconds have passed since we received the last update from the hierarchical controller - to avoid the execution of multiple optimizations.
        tnow = calendar.timegm(time.gmtime())

        if tnow - self.__timelastjob < 2:
            # Do not forget to release!!!
            self.__lock.release()
            return False

        for job in self.__statflowjobs:

            # Add a static flow
            if job['TYPE'] == 'ADD':
                flow = self.getFlowExEx(job['PROTO'], job['SIP'], job['DIP'])
                if flow == None:
                    print "*** WARNING: unable to find flow based on: ", job['PROTO'], job['SIP'], job['DIP'], "(it could be a reverse flow, so no need to configure it)"
                    continue

                # At this point we have found the flow, we need to check if the ACCESS/EGGRESS nodes/ports have changed
                # Get the mapping to the real nodes and ports based on the emulated switch's configuration
                in_nodeID = mod_emswitch.of_emsw.ports[job['INP']][ModEMSwitch.REALSW]
                out_nodeID = mod_emswitch.of_emsw.ports[job['OUTP']][ModEMSwitch.REALSW]
                in_port = mod_emswitch.of_emsw.ports[job['INP']][ModEMSwitch.REALP]
                out_port = mod_emswitch.of_emsw.ports[job['OUTP']][ModEMSwitch.REALP]

                if flow.getProp(ModFlow.ACC) != in_nodeID or flow.getProp(ModFlow.ACCPORT) != in_port:
                    flow.addProp(ModFlow.ACC, in_nodeID)
                    flow.addProp(ModFlow.ACCPORT, in_port)
                    flow.accChanged = True
                    print ("*** Access end-point changed for flow. New configuration: ", flow.getName(), flow.getProp(ModFlow.ACC), flow.getProp(ModFlow.EGG), 
                                                                flow.getProp(ModFlow.ACCPORT), flow.getProp(ModFlow.EGGPORT))
                    retVal = True
                if flow.getProp(ModFlow.EGG) != out_nodeID or flow.getProp(ModFlow.EGGPORT) != out_port:
                    flow.addProp(ModFlow.EGG, out_nodeID)
                    flow.addProp(ModFlow.EGGPORT, out_port)
                    flow.eggChanged = True
                    print ("*** Egress end-point changed for flow. New configuration: ", flow.getName(), flow.getProp(ModFlow.ACC), flow.getProp(ModFlow.EGG), 
                                                                flow.getProp(ModFlow.ACCPORT), flow.getProp(ModFlow.EGGPORT))
                    retVal = True
            else:
                # Remove a static flow (we do not need to delete it from the internal configuration, its associated flows will be deleted by the controller)
                pass

        # Reset list
        self.__statflowjobs = []        

        self.__lock.release()

        return retVal
