# Copyright 2016 Bela Genge
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
OptimalFlow/mod_setup/ModSetup

@author: Bela Genge
'''

from mod_link import ModLink
from mod_flow import ModFlow
from mod_emswitch import ModEMSwitch
import re

class ModSetup():
    '''
    '''

    def __init__(self, fname = './setup.ini'):
        '''
        Constructor.
        '''
        self.__scippath = ''
        self.__fname = fname
        self.__nodes = []
        self.__links = []
        self.__flows = []
        self.__nodep = []
        self.__linkp = []
        self.__flowp = []
        self.__secur = [] # Security properties
        self.__backup = [] # back-up lines

        # Emulated switch stuff
        self.__emsid = ''
        self.__emschost = ''
        self.__emscport = ''
        self.__emsports = []
        self.__emsprop = []
        self.__emsused = 'No'


    def readSetup(self):
        '''
        Read the setup file and extract data. Each link needs to start either with "Nodes: " or "Link: "
        '''
        with open(self.__fname, 'r') as f:
            for line in f:
                # Skip empty lines
                if len(line) <= 0:
                    continue

                # Skip comments
                if line[0] == '#':
                    continue

                line = line.rstrip('\r\n')
                if re.search('SCIPPath> ', line):
                    self.__scippath = line.split('>')[1].strip(' ')
                elif re.search('Nodes> ', line):
                    self.__nodes.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('Security> ', line):
                    self.__secur = line.split('>')[1].strip(' ').split(' ')
                elif re.search('BackupLines> ', line):
                    self.__backup = line.split('>')[1].strip(' ').split(' ')
                elif re.search('Link> ', line):
                    self.__links.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('Flow> ', line):
                    self.__flows.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('NodeProp> ', line):
                    self.__nodep.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('LinkProp> ', line):
                    self.__linkp.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('FlowProp> ', line):
                    self.__flowp.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('SWID> ', line):
                    self.__emsid = line.split('>')[1].strip(' ')
                elif re.search('SWCtHost> ', line):
                    self.__emschost = line.split('>')[1].strip(' ')
                elif re.search('SWCtPort> ', line):
                    self.__emscport = line.split('>')[1].strip(' ')
                elif re.search('SWPorts> ', line):
                    self.__emsports.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('SWProp> ', line):
                    self.__emsprop.append(line.split('>')[1].strip(' ').split(' '))
                elif re.search('SWUsed> ', line):
                    self.__emsused = line.split('>')[1].strip(' ')

        print "********** SETUP FILE ************"
        print "SCIPPath: "
        print self.__scippath
        print "Nodes: "
        print self.__nodes
        print "Security: "
        print self.__secur
        print "Backup lines: "
        print self.__backup
        print "Links: "
        print self.__links
        print "Flows: "
        print self.__flows
        print "NodeProp: "
        print self.__nodep
        print "LinkProp: "
        print self.__linkp
        print "FlowProp: "
        print self.__flowp
        print "SWID: ", self.__emsid
        print "SWCtHost: ", self.__emschost
        print "SWCtPort: ", self.__emscport
        print "SWPorts: "
        print self.__emsports
        print "SWProp: "
        print self.__emsprop
        print "SWUsed: "
        print self.__emsused


    def getSCIPPath(self):
        return self.__scippath


    def getNodes(self):
        return self.__nodes[0]


    def getLinks(self):
        return self.__links


    def getFlows(self):
        return self.__flows


    def getNodeProp(self, node = ''):
        if node == '':
            return self.__nodep

        np = {}
        for prop in self.__nodep:
            if prop[0] == node:
                np[prop[1]] = prop[2]
        return np


    def getLinkProp(self, link = ''):
        if link == '':
            return self.__linkp

        # Add nodes
        lp = {}
        for ll in self.__links:
            if ll[0] == link:
                lp[ModLink.N1] = ll[1]
                lp[ModLink.N2] = ll[2]

        # Add other properties
        for ll in self.__linkp:
            if ll[0] == link:
                if ll[1] == ModLink.CAP:
                    lp[ll[1]] = int(ll[2])
                else:
                    lp[ll[1]] = ll[2]

        return lp


    def getFlowProp(self, flow = ''):
        if flow == '':
            return self.__flowp

        # Add access/egress nodes
        fp = {}
        for ff in self.__flows:
            if ff[0] == flow:
                fp[ModFlow.ACC] = ff[1]
                fp[ModFlow.EGG] = ff[2]

        # Add other properties
        for ff in self.__flowp:
            if ff[0] == flow:
                if ff[1] == ModFlow.DEM:
                    fp[ff[1]] = int(ff[2])
                else:
                    fp[ff[1]] = ff[2]

        return fp


    def getEmSwitchID(self):
        return self.__emsid


    def getEmSwitchCtHost(self):
        return self.__emschost


    def getEmSwitchCtPort(self):
        return self.__emscport


    def getEmSwitchPorts(self):
        return self.__emsports[0]


    def getEmSwitchProp(self, port):

        sp = {}
        for prop in self.__emsprop:
            if prop[0] == port:
                sp[prop[1]] = prop[2]

        return sp


    def getEmSwitchUsed(self):
        return self.__emsused


    def getLinkSecurProp(self, link):

        sp = []
        for ll in self.__linkp:
            if ll[0] == link and ll[1] == ModLink.SEC:
                sp.append(ll[2])

        return sp


    def getFlowSecurProp(self, flow):

        sp = []
        for ff in self.__flowp:
            if ff[0] == flow and ff[1] == ModFlow.SEC:
                sp.append(ff[2])

        return sp


    def getSecurProp(self):

        return self.__secur

    
    def getBackupLinks(self):

        return self.__backup
